/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 *
 */

#include <LLMJVM_impl.h>

#include "microej_time.h"
#include "gina.h"
#include "tx_api.h"
#include "tx_semaphore.h"
#include "qapi_timer.h"

#define MILLISECONDS_BY_TICK (10)

// Uncomment to print debug informations.
// #define DEBUG
#ifdef DEBUG
#define DEBUG_LOG_INFO gina_uwlog_printf("[LLMJVM][INFO]: %s:", __func__);\
							gina_uwlog_printf
#define DEBUG_LOG_ERROR gina_uwlog_printf("[LLMJVM][ERROR]: %s:", __func__);\
							gina_uwlog_printf
#else
#define DEBUG_LOG_INFO(...) ((void) 0)
#define DEBUG_LOG_ERROR(...) ((void) 0)
#endif // DEBUG

static TX_SEMAPHORE* idle_semaphore;
static qapi_TIMER_handle_t schedule_timer;
static int64_t application_time_offset = 0;
static int64_t current_schedule_time = INT64_MAX;

/**
 * @brief The callback for the schedule request timer.
 * 
 * @param param Parameter passed during timer creation.
 */
static void schedule_request_callback(uint32_t param) {
	DEBUG_LOG_INFO("");
	LLMJVM_schedule();
}

/**
 * This function is called once during MicroJvm virtual machine Startup. It may be used to initialize specific data.<br>
 * @return {@link LLMJVM_OK} on success, {@link LLMJVM_ERROR} on error.
 */
int32_t LLMJVM_IMPL_initialize(void) {
	uint8_t result = 0;
	qapi_Status_t status = 0;
	qapi_TIMER_define_attr_t schedule_timer_define_attr = {0};

	/* Allocate semaphore */
	result = txm_module_object_allocate((void*) &idle_semaphore, sizeof(TX_SEMAPHORE));
	if (TX_SUCCESS != result) {
		DEBUG_LOG_ERROR("txm_module_object_allocate returned %d", result);
		return LLMJVM_ERROR;
	}

	/* Create an idle semaphore */
	result = tx_semaphore_create(idle_semaphore, "Idle semaphore", 1);
	if (TX_SUCCESS != result) {
		DEBUG_LOG_ERROR("Error creating semaphore, result=0x%x", result);
		return LLMJVM_ERROR;
	}

	/* Define a timer for the schedule request */
	schedule_timer_define_attr.cb_type = QAPI_TIMER_FUNC1_CB_TYPE;
	schedule_timer_define_attr.sigs_func_ptr = schedule_request_callback;
	schedule_timer_define_attr.sigs_mask_data = 0;
	schedule_timer_define_attr.deferrable = TRUE;
	status = qapi_Timer_Def(&schedule_timer, &schedule_timer_define_attr);
	if (QAPI_OK != status) {
		DEBUG_LOG_ERROR("Error defining timer, status=%d", status);
		return LLMJVM_ERROR;
	}

	return LLMJVM_OK;
}

/**
 * This function is called once during the MicroJvm virtual machine initialization by the MicroJvm virtual machine task.
 * It can be useful if the MicroJvm virtual machine support needs to know the MicroJvm virtual machine task.
 * @return {@link LLMJVM_OK} on success, {@link LLMJVM_ERROR} on error.
 */
int32_t LLMJVM_IMPL_vmTaskStarted(void) {
	return LLMJVM_OK;
}

/**
 * Schedule an alarm (or timer) that will be triggered at the given absolute (system) time.
 * If an alarm is already scheduled for an earlier time this function must do nothing,
 * otherwise it must configure the alarm. If the given absolute time has already been reached, this
 * function must call {@link LLMJVM_schedule}.
 * The scheduled alarm must call the function {@link LLMJVM_schedule} when it is triggered.
 * The specified time is in milliseconds.
 * @return {@link LLMJVM_OK} on success, {@link LLMJVM_ERROR} on error.
 */
int32_t LLMJVM_IMPL_scheduleRequest(int64_t absolute_time_milliseconds) {
	DEBUG_LOG_INFO("");
	qapi_Status_t status = 0;
	int64_t relative_time_milliseconds = 0;
	qapi_TIMER_set_attr_t schedule_timer_set_attr = {0};

	/* Check if the requested time is lower than the current schedule time */
	if (absolute_time_milliseconds < current_schedule_time) {
		/* Save the new alarm schedule time */
		current_schedule_time = absolute_time_milliseconds;

		/* Stop the schedule timer */
		qapi_Timer_Stop(schedule_timer);

		relative_time_milliseconds = absolute_time_milliseconds - microej_timer_get_current_time();

		if (0 < relative_time_milliseconds) {
			/* Update the timer and launch it*/
			schedule_timer_set_attr.reload = FALSE;
			schedule_timer_set_attr.unit = QAPI_TIMER_UNIT_MSEC;
			schedule_timer_set_attr.time = relative_time_milliseconds;
			status = qapi_Timer_Set(schedule_timer, &schedule_timer_set_attr);
			if (QAPI_OK != status) {
				DEBUG_LOG_ERROR("Error updating timer, status=%d", status);
				return LLMJVM_ERROR;
			}
			DEBUG_LOG_INFO("Timer started");
			return LLMJVM_OK;
		} else {
			/* Absolute time has already been reached */
			return LLMJVM_schedule();
		}
	}

	return LLMJVM_OK;
}

/**
 * Causes the MicroJvm virtual machine RTOS task to sleep until it is woken up by the {@link LLMJVM_wakeupVM} function.
 * This function is called by the MicroJvm virtual machine task.
 * @return {@link LLMJVM_OK} if wakeup occurred, {@link LLMJVM_INTERRUPTED} if the MicroJvm virtual machine task
 * has been interrupted, or {@link LLMJVM_ERROR} on error.
 */
int32_t LLMJVM_IMPL_idleVM(void) {
	DEBUG_LOG_INFO("");
	uint8_t result = 0;

	/* Try to take the idle semaphore */
	result = tx_semaphore_get(idle_semaphore, TX_WAIT_FOREVER);
	if (TX_SUCCESS != result) {
		DEBUG_LOG_ERROR("Error getting semaphore, result=0x%x", result);
		return LLMJVM_ERROR;
	}

	return LLMJVM_OK;
}

/**
 * Wake up the MicroJvm virtual machine RTOS task. If the MicroJvm virtual machine task is not sleeping,
 * the wakeup stays pending and the MicroJvm virtual machine will not sleep on the next {@link LLMJVM_idleVM} call unless
 * there is a call to {@link LLMJVM_ackWakeup} between this call and the next {@link LLMJVM_idleVM} call.
 * This function must be called only by the MicroJvm virtual machine code. If a task wants to wake up the MicroJvm
 * virtual machine, it must use the {@link LLMJVM_schedule} function (which may in turn call
 * this function).
 * @return {@link LLMJVM_OK} on success, {@link LLMJVM_ERROR} on error.
 */
int32_t LLMJVM_IMPL_wakeupVM(void) {
	DEBUG_LOG_INFO("");
	uint8_t result = 0;

	/* Reset the schedule time */
	current_schedule_time = INT64_MAX;

	/* Release the semaphore */
	result = tx_semaphore_put(idle_semaphore);
	if (TX_SUCCESS != result) {
		DEBUG_LOG_ERROR("Error putting semaphore, result=0x%x", result);
		return LLMJVM_ERROR;
	}

	return LLMJVM_OK;
}

/**
 * Clears any outstanding {@link LLMJVM_wakeupVM} request. After calling this function, a call to {@link LLMJVM_idleVM} will
 * result in a wait even if {@link LLMJVM_wakeupVM} has been called previously (provided no other {@link LLMJVM_wakeupVM} call
 * has occurred since the call to this function.
 * This function is called by the MicroJvm virtual machine task.
 * @return {@link LLMJVM_OK} on success, {@link LLMJVM_ERROR} on error.
 */
int32_t LLMJVM_IMPL_ackWakeup(void) {
	return LLMJVM_OK;
}

/**
 * Returns the ID of the current task.
 */
int32_t LLMJVM_IMPL_getCurrentTaskID(void) {
	return (int32_t) tx_thread_identify();
}

/**
 * This function is called during MicroJvm virtual machine end. It may be used to freed specific data.<br>
 * @return {@link LLMJVM_OK} on success, {@link LLMJVM_ERROR} on error.
 */
int32_t LLMJVM_IMPL_shutdown(void) {
	return LLMJVM_OK;
}

/**
 * Sets the application time. The application time is the difference, measured in milliseconds,
 * between the current time and midnight, January 1, 1970 UTC.
 * This time does not change the system time.
 * @param t the application time to set in milliseconds
 */
void LLMJVM_IMPL_setApplicationTime(int64_t t) {
	int64_t current_time = microej_timer_get_current_time();
	application_time_offset = t - current_time;
}

/**
 * Gets the system or the application time in milliseconds.
 * The system time is the running time since the start of the device. This time is independent from any user considerations
 * and cannot be changed.
 * The application time is the difference, measured in milliseconds, between the current time and midnight, January 1, 1970 UTC.
 * @param system if 1, get the system time, otherwise get the application time.
 */
int64_t LLMJVM_IMPL_getCurrentTime(uint8_t system) {
	int64_t system_time = microej_timer_get_current_time();
	return system ? system_time : system_time + application_time_offset;
}

/**
 * Gets the current timestamp in nanoseconds.
 * Only elapsed time between two calls is meaningful.
 * @return the current timestamp in nanoseconds
 */
int64_t LLMJVM_IMPL_getTimeNanos(void) {
	return microej_timer_get_time_nanos();
}
