/**
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */

#include <LLAT_impl.h>
#include "gina.h"
#include "sni.h"
#include "osal.h"
#include "microej_main.h"

#ifdef __cplusplus
extern "C" {
#endif

// #define LLAT_DEBUG
#ifdef LLAT_DEBUG
#define LLAT_DEBUG_INFO gina_uwlog_printf
#define LLAT_DEBUG_ERROR gina_uwlog_printf
#else
#define LLAT_DEBUG_INFO(...) ((void) 0)
#define LLAT_DEBUG_ERROR(...) ((void) 0)
#endif // LLAT_DEBUG

static GinaIoHndl handle = -1;
static jboolean data_available = true;
static int32_t waiting_thread = -1;
static OSAL_mutex_handle_t mutex;

static void AT_cleanup(void* param);
static void AT_on_data_available(GinaIoHndl hdl, void *data);

static gstatus AT_init(void) {
	LLAT_DEBUG_INFO("%s", __func__);
	gstatus status;
	OSAL_status_t osal_status;
	GinaIoHndl gina_io_hndl;

	gina_io_hndl = gina_atc_open(GINA_ATC1, &status);
	if (GINA_ERROR_EXISTS == status) {
		// Channel 1 is already opened, try with channel 2
		gina_io_hndl = gina_atc_open(GINA_ATC2, &status);
	}
	LLAT_DEBUG_INFO("%s: gina_atc_open() returned %d, status=%d", __func__, gina_io_hndl, status);

	if (GINA_NOERROR == status) {
		handle = gina_io_hndl;
		add_cleanup_callback(AT_cleanup);
		LLAT_DEBUG_INFO("%s: add_cleanup_callback()", __func__); // TODO remove

		status = gina_atc_ioctl(handle, GINA_ATC_IOCTL_CB_REG, AT_on_data_available, NULL);
		if (GINA_NOERROR != status) {
			LLAT_DEBUG_ERROR("%s: gina_atc_ioctl() failed with status=%d", __func__, status);
			return status;
		}

		osal_status = OSAL_mutex_create((uint8_t*) "LLAT_mutex", &mutex);
		if (OSAL_OK != osal_status) {
			LLAT_DEBUG_ERROR("%s: OSAL_mutex_create() failed with status=%d", __func__, osal_status);
			return GINA_ERROR;
		}
	}

	return status;
}

static void AT_cleanup(void* param) {
	LLAT_DEBUG_INFO("%s", __func__);
	gstatus status;
	OSAL_status_t osal_status;

	if (0 > handle) { // Nothing to clean
		LLAT_DEBUG_INFO("%s: nothing to clean", __func__);
		return;
	}

	status = gina_atc_ioctl(handle, GINA_ATC_IOCTL_CB_UNREG, NULL, NULL);
	if (GINA_NOERROR != status) {
		LLAT_DEBUG_ERROR("%s: gina_atc_ioctl() failed with status=%d", __func__, status);
	}

	osal_status = OSAL_mutex_delete(&mutex);
	if (OSAL_OK != osal_status) {
		LLAT_DEBUG_ERROR("%s: OSAL_mutex_delete() failed with status=%d", __func__, osal_status);
	}

	status = gina_atc_close(handle);
	if (GINA_NOERROR == status) {
		handle = 0;
	} else {
		LLAT_DEBUG_ERROR("%s: gina_atc_close() failed with status=%d", __func__, status);
	}
}

/**
 * @brief Read data from the AT connection.
 *
 * @param data Destination buffer for the received data.
 * @param length Length of the data to read.
 * @return The length of data read, otherwise a negative value.
 */
jint LLAT_read_data(jbyte* data, jint length) {
	LLAT_DEBUG_INFO("%s", __func__);
	gstatus status;
	gint32 read;
	OSAL_status_t osal_status;

	if (0 > handle) {
		status = AT_init();
		if (GINA_NOERROR != status) {
			LLAT_DEBUG_ERROR("%s: AT init failed with status: %d", __func__, status);
			return -1;
		}
	}

	osal_status = OSAL_mutex_take(&mutex, OSAL_INFINITE_TIME);
	if (OSAL_OK != osal_status) {
		LLAT_DEBUG_ERROR("%s: OSAL_mutex_take() failed with status=%d", __func__, osal_status);
		return -1;
	}

	read = gina_atc_read(handle, data, length, &status);

	if (GINA_NOERROR != status) {
		LLAT_DEBUG_ERROR("%s: Read failed with status: %d", __func__, status);
		osal_status = OSAL_mutex_give(&mutex);
		if (OSAL_OK != osal_status) {
			LLAT_DEBUG_ERROR("%s: OSAL_mutex_give() failed with status=%d", __func__, osal_status);
		}
		return -1;
	}

	if (0 == read) {
		data_available = false;
	}

	osal_status = OSAL_mutex_give(&mutex);
	if (OSAL_OK != osal_status) {
		LLAT_DEBUG_ERROR("%s: OSAL_mutex_give() failed with status=%d", __func__, osal_status);
	}

	return read;
}

/**
 * @brief Send data over the AT connection.
 *
 * @param data Buffer of data to send.
 * @param length Length of the buffer.
 * @return A negative value if the data failed to be sent.
 */
jint LLAT_send_data(jbyte* data, jint length) {
	LLAT_DEBUG_INFO("%s", __func__);
	gstatus status;
	gint32 written;

	if (0 > handle) {
		status = AT_init();
		if (GINA_NOERROR != status) {
			LLAT_DEBUG_ERROR("%s: AT init failed with status=%d", __func__, status);
			return -1;
		}
	}

	written = gina_atc_write(handle, data, length-1, &status); // data is null terminated

	if (GINA_NOERROR != status) {
		LLAT_DEBUG_ERROR("%s: AT write failed with status=%d", __func__, status);
		return -1;
	}

	return written;
}

/**
 * @brief Wait for data on the AT connection.
 *
 */
void LLAT_wait_for_data(void) {
	LLAT_DEBUG_INFO("%s", __func__);
	gstatus gina_status;
	int32_t sni_status, java_thread_id;
	OSAL_status_t osal_status;

	if (0 > handle) {
		LLAT_DEBUG_INFO("%s: AT channel is not initialized", __func__);
		return;
	}

	java_thread_id = SNI_getCurrentJavaThreadID();
	if (SNI_ERROR == java_thread_id) {
		LLAT_DEBUG_INFO("%s: Could not get Java thread ID", __func__);
		return; // Must be called from a Java thread
	}

	osal_status = OSAL_mutex_take(&mutex, OSAL_INFINITE_TIME);
	if (OSAL_OK != osal_status) {
		LLAT_DEBUG_ERROR("%s: OSAL_mutex_take() failed with status=%d", __func__, osal_status);
		return;
	}
	
	if (false == data_available && // no data available
		     0 > waiting_thread) { // no thread already waiting for data
		waiting_thread = java_thread_id;

		sni_status = SNI_suspendCurrentJavaThread(0);
		if (SNI_ERROR == sni_status) {
			LLAT_DEBUG_ERROR("%s: Could not get Java thread ID", __func__);
		}
	}

	osal_status = OSAL_mutex_give(&mutex);
	if (OSAL_OK != osal_status) {
		LLAT_DEBUG_ERROR("%s: OSAL_mutex_give() failed with status=%d", __func__, osal_status);
	}
}

static void AT_on_data_available(GinaIoHndl hdl, void *data) {
	LLAT_DEBUG_INFO("%s", __func__);
	int32_t sni_status;
	OSAL_status_t osal_status;

	osal_status = OSAL_mutex_take(&mutex, OSAL_INFINITE_TIME);
	if (OSAL_OK != osal_status) {
		LLAT_DEBUG_ERROR("%s: OSAL_mutex_take() failed with status=%d", __func__, osal_status);
		return;
	}

	data_available = true;

	if (0 <= waiting_thread) {
		sni_status = SNI_resumeJavaThread(waiting_thread);
		if (SNI_ERROR == sni_status) {
			LLAT_DEBUG_ERROR("%s: Could not resume Java thread", __func__);
		} else {
			waiting_thread = -1;
		}
	}

	osal_status = OSAL_mutex_give(&mutex);
	if (OSAL_OK != osal_status) {
		LLAT_DEBUG_ERROR("%s: OSAL_mutex_give() failed with status=%d", __func__, osal_status);
	}
}

#ifdef __cplusplus
}
#endif
