/**
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */

#include <LLPDP_impl.h>
#include "qapi_dss.h"
#include "qapi_timer.h"
#include "tx_event_flags.h"
#include "gina.h"
#include "qapi_dnsc.h"
#include "LLNET_DNS_exs82.h"

#ifdef __cplusplus
extern "C" {
#endif

#define EVENT_CONNECTED (1)

static TX_EVENT_FLAGS_GROUP* netctrl_event_flags;

qapi_DSS_Hndl_t netctrl_dss_handle;

/* Receives network event. This function is usually called into a syscall : do not call any gina/qapi functions. */
static void netctrl_event_cb (qapi_DSS_Hndl_t hndl, void* user_data, qapi_DSS_Net_Evt_t evt, qapi_DSS_Evt_Payload_t* payload_ptr) {
	switch (evt) {
	case QAPI_DSS_EVT_NET_IS_CONN_E:
		if (TX_SUCCESS != tx_event_flags_set(netctrl_event_flags, EVENT_CONNECTED, TX_OR)) {
			gina_uwlog_printf("%s: Failed to set flag CONNECTED", __func__);
		}
		break;
	case QAPI_DSS_EVT_NET_NO_NET_E:
		break;
	case QAPI_DSS_EVT_NET_RECONFIGURED_E:
		break;
	case QAPI_DSS_EVT_NET_NEWADDR_E:
		break;
	case QAPI_DSS_EVT_NET_DELADDR_E:
		break;
	case QAPI_DSS_EVT_NIPD_DL_DATA_E:
		break;
	default:
		// Unknown event
		break;
	}
}

/**
 * @brief Activate a session with the given configuration in parameters.
 *
 * @param apn Buffer for the APN.
 * @param apn_length Length of the APN buffer.
 * @param login Buffer for the login.
 * @param login_length Length of the login buffer.
 * @param pwd Buffer for the secret.
 * @param pwd_length Length for the secret buffer.
 * @return The ID of the session, a negative value otherwise.
 */
jint LLPDP_activate(jbyte* apn, jint apn_length, jbyte* login, jint login_length, jbyte* pwd, jint pwd_length) {
	qapi_Status_t status;
	qapi_DSS_Hndl_t handle;
	qapi_DSS_Call_Param_Value_t value;
	
	// Allocate an event flag for the event callback
	if (TX_SUCCESS != txm_module_object_allocate((void*) &netctrl_event_flags, sizeof(TX_EVENT_FLAGS_GROUP))) {
		gina_uwlog_printf("%s: Failed to allocate event flags", __func__);
		return -1;
	}

	// Create an event flag for the event callback
	if (TX_SUCCESS != tx_event_flags_create(netctrl_event_flags, "Net Ctrl Events")) {
		gina_uwlog_printf("%s: Failed to create event flags", __func__);
		return -1;
	}

	do {
		status = qapi_DSS_Get_Data_Srvc_Hndl(netctrl_event_cb, NULL, &handle);
		if (QAPI_OK != status) {
			// TODO : retry every 500ms on failure (because it may fail at start up)
			gina_uwlog_printf("%s: qapi_DSS_Get_Data_Srvc_Hndl() failed with status %d", __func__, status);
			return -1;
		}
	} while (NULL == handle);
	netctrl_dss_handle = handle;

	// Technology preference (UMTS, CDMA, 1X, DO, LTE or TDSCDMA)
	value.buf_val = NULL;
	value.num_val = QAPI_DSS_RADIO_TECH_UNKNOWN; // Auto
	status = qapi_DSS_Set_Data_Call_Param(handle, QAPI_DSS_CALL_INFO_TECH_PREF_E, &value);
	if (QAPI_OK != status) {
		gina_uwlog_printf("%s: qapi_DSS_Set_Data_Call_Param() for \"Technology preference\" failed with status %d", __func__, status);
		return -1;
	}

	// APN
	value.buf_val = (char*) apn;
	value.num_val = apn_length;
	status = qapi_DSS_Set_Data_Call_Param(handle, QAPI_DSS_CALL_INFO_APN_NAME_E, &value);
	if (QAPI_OK != status) {
		gina_uwlog_printf("%s: qapi_DSS_Set_Data_Call_Param() for \"APN\" failed with status %d", __func__, status);
		return -1;
	}

	// Username
	value.buf_val = (char*) login;
	value.num_val = login_length;
	status = qapi_DSS_Set_Data_Call_Param(handle, QAPI_DSS_CALL_INFO_USERNAME_E, &value);
	if (QAPI_OK != status) {
		gina_uwlog_printf("%s: qapi_DSS_Set_Data_Call_Param() for \"Username\" failed with status %d", __func__, status);
		return -1;
	}

	// Password
	value.buf_val = (char*) pwd;
	value.num_val = pwd_length;
	status = qapi_DSS_Set_Data_Call_Param(handle, QAPI_DSS_CALL_INFO_PASSWORD_E, &value);
	if (QAPI_OK != status) {
		gina_uwlog_printf("%s: qapi_DSS_Set_Data_Call_Param() for \"Password\" failed with status %d", __func__, status);
		return -1;
	}

	// IP version
	value.buf_val = NULL;
	value.num_val = QAPI_DSS_IP_VERSION_4;
	status = qapi_DSS_Set_Data_Call_Param(handle, QAPI_DSS_CALL_INFO_IP_VERSION_E, &value);
	if (QAPI_OK != status) {
		gina_uwlog_printf("%s: qapi_DSS_Set_Data_Call_Param() for \"IP version\" failed with status %d", __func__, status);
		return -1;
	}

	// Profile index (configured using the "+CGDCONT" AT command)
	value.buf_val = NULL;
	value.num_val = 1; // TODO configure the PDP context
	status = qapi_DSS_Set_Data_Call_Param(handle, QAPI_DSS_CALL_INFO_UMTS_PROFILE_IDX_E, &value);
	if (QAPI_OK != status) {
		gina_uwlog_printf("%s: qapi_DSS_Set_Data_Call_Param() for \"Profile index\" failed with status %d", __func__, status);
		return -1;
	}
	
	status = qapi_DSS_Start_Data_Call(handle);
	if (QAPI_OK != status) {
		gina_uwlog_printf("%s: qapi_DSS_Start_Data_Call() failed with status %d", __func__, status);
		return -1;
	}

	uint32_t actual_flags;
	if (TX_SUCCESS != tx_event_flags_get(netctrl_event_flags, EVENT_CONNECTED, TX_OR, &actual_flags, 5000)) {
		gina_uwlog_printf("%s: Could not retrieve event", __func__);
		return -1;
	}

	if(EVENT_CONNECTED & actual_flags) {
		LLNET_DNS_addDnsServer(handle);
		return (jint) handle;
	}

	gina_uwlog_printf("%s: Not connected", __func__);
	return -1;
}

/**
 * @brief Deactivate the session.
 *
 * @param id The ID of the session to deactivate.
 * @return A negative value if the session failed to be deactivated.
 */
jint LLPDP_deactivate(jint id) {
	qapi_Status_t status;
	qapi_DSS_Hndl_t handle;

	handle = (qapi_DSS_Hndl_t) id;

	status = qapi_DSS_Stop_Data_Call(handle);
	if (QAPI_OK != status) {
		return -1;
	}

	// TODO wait for the tear-down event

	return 0;
}

/**
 * @brief Returns the APN maximum size.
 *
 * @return jint
 */
jint LLPDP_get_apn_max_size(void) {
	return QAPI_DSS_CALL_INFO_APN_MAX_LEN;
}

/**
 * @brief Returns the login maximum size.
 *
 * @return jint
 */
jint LLPDP_get_login_max_size(void) {
	return QAPI_DSS_CALL_INFO_USERNAME_MAX_LEN;
}

/**
 * @brief Returns the password maximum size.
 *
 * @return jint
 */
jint LLPDP_get_pwd_max_size(void) {
	return QAPI_DSS_CALL_INFO_PASSWORD_MAX_LEN;
}

/**
 * @brief Method that blocks until the session is deactivated.
 *
 * @param The ID of the session.
 */
void LLPDP_wait_deactivation(jint id) {
	// TODO I think it would be better to change the foundation here : use a pump to forward event generated by the qapi_DSS_Net_Ev_CB_t callback (netctrl_event_cb).
	// With this change, we would be able not only to wait for deactivation but also get connection, disconnection, IP address assignment/deassignment events (see qapi_DSS_Net_Evt_t).
}

#ifdef __cplusplus
}
#endif
