/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 *
 */

#include <LLGNSS_impl.h>
#include "gina.h"
#include "qapi_location.h"
#include "sni.h"
#include "osal.h"
#include "microej_main.h"

#define USER_BUFFER_SIZE 1024

#define ERROR_GNSS_INITIALIZATION   1
#define ERROR_GNSS_TRACKING         2
#define ERROR_GNSS_CLOSE            3

// Uncomment to print debug informations.
// #define DEBUG_GNSS
#ifdef DEBUG_GNSS
#define DEBUG_GNSS_LOG_INFO(...) gina_uwlog_printf("[GNSS][INFO]: " __VA_ARGS__);
#define DEBUG_GNSS_LOG_ERROR(...) gina_uwlog_printf("[GNSS][ERROR]: " __VA_ARGS__);
#else
#define DEBUG_GNSS_LOG_INFO(...) ((void) 0)
#define DEBUG_GNSS_LOG_ERROR(...) ((void) 0)
#endif // DEBUG_GNSS

static qapi_loc_client_id client_id;             // Location module session ID
static uint32_t tracking_id;                     // Position tracking session ID
static qapi_Location_t current_location;         // Last fix data
static OSAL_binary_semaphore_handle_t semaphore; // Semaphore used to wait for fix callback

static uint8_t userBuffer[USER_BUFFER_SIZE];

static void GNSS_cleanup(void* param);

static void capabilities_callback(qapi_Location_Capabilities_Mask_t mask) {
	DEBUG_GNSS_LOG_INFO("GNSS capabilities :");
	if (mask & QAPI_LOCATION_CAPABILITIES_TIME_BASED_TRACKING_BIT) {
		DEBUG_GNSS_LOG_INFO("\tTime based tracking");
	}
	if (mask & QAPI_LOCATION_CAPABILITIES_TIME_BASED_BATCHING_BIT) {
		DEBUG_GNSS_LOG_INFO("\tTime based batching");
	}
	if (mask & QAPI_LOCATION_CAPABILITIES_DISTANCE_BASED_TRACKING_BIT) {
		DEBUG_GNSS_LOG_INFO("\tDistance based tracking");
	}
	if (mask & QAPI_LOCATION_CAPABILITIES_DISTANCE_BASED_BATCHING_BIT) {
		DEBUG_GNSS_LOG_INFO("\tDistance based batching");
	}
	if (mask & QAPI_LOCATION_CAPABILITIES_GEOFENCE_BIT) {
		DEBUG_GNSS_LOG_INFO("\tGeofence");
	}
	if (mask & QAPI_LOCATION_CAPABILITIES_GNSS_DATA_BIT) {
		DEBUG_GNSS_LOG_INFO("\tGNSS data");
	}
}

static void response_callback(qapi_Location_Error_t error, uint32_t id) {
	if (QAPI_LOCATION_ERROR_SUCCESS != error) {
		DEBUG_GNSS_LOG_ERROR("Request failed (id=%d, err=%d)", id, error);
	}
}

static void collective_response_callback(size_t count, qapi_Location_Error_t* error, uint32_t* ids) {
	for (int i = 0; i < count; i++) {
		response_callback(error[i], ids[i]);
	}
}

static void tracking_callback(qapi_Location_t location) {
	DEBUG_GNSS_LOG_INFO("Location event received (tracking)");
	memcpy(&current_location, &location, sizeof(qapi_Location_t));
}

static void singleshot_callback(qapi_Location_t location, qapi_Location_Error_t error) {
	OSAL_status_t osal_status = 0;

	if (QAPI_LOCATION_ERROR_SUCCESS == error) {
		DEBUG_GNSS_LOG_INFO("Location event received (singleshot)");
		memcpy(&current_location, &location, sizeof(qapi_Location_t));
	} else {
		DEBUG_GNSS_LOG_ERROR("Localisation failed (err=%d)", error);
	}

	osal_status = OSAL_binary_semaphore_give(&semaphore);
	if (OSAL_OK != osal_status) {
		DEBUG_GNSS_LOG_ERROR("OSAL_binary_semaphore_give failed (err=%d)", osal_status);
	}
}

/**
 * @brief Gets the latitude in degrees.
 * 
 * @return The latitude in degrees
 */
float LLGNSS_getLatitude(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);
	return (float) current_location.latitude;
}

/**
 * @brief Gets the longitude in degrees.
 * 
 * @return The longitude in degrees.
 */
float LLGNSS_getLongitude(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);
	return (float) current_location.longitude;
}

/**
 * @brief Gets the altitude in meters.
 * 
 * @return The altitude in meters.
 */
float LLGNSS_getAltitude(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);
	return (float) current_location.altitude;
}

/**
 * @brief Gets the velocity in meters per second.
 * 
 * @return The velocity in meters per second.
 */
float LLGNSS_getVelocity(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);
	return (float) current_location.speed;
}

/**
 * @brief Gets the time elapsed since the Epoch in milliseconds.
 * This is the time corresponding to the positioning data and
 * it should not be used as the system time.
 * 
 * @return The time in milliseconds.
 */
int64_t LLGNSS_getTime(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);
	return (int64_t) current_location.timestamp;
}

/**
 * @brief Fetches and update the positioning data.
 * This function refresh the data but does not actually return it,
 * in order to retrieve positioning data, get() functions still needs to be called.
 * 
 */
void LLGNSS_fetch(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);

	qapi_Location_Error_t error = 0;
	qapi_Singleshot_Options_t options = {0};
	uint32_t session_id = 0;
	OSAL_status_t osal_status = 0;
	
	options.size = sizeof(qapi_Singleshot_Options_t);
	// options.accuracyLevel = ;
	// options.powerLevel = ;
	options.timeout = 30000;

	error = qapi_Loc_Get_Single_Shot_V2(client_id, &options, &session_id);
	if (QAPI_LOCATION_ERROR_SUCCESS != error) {
		DEBUG_GNSS_LOG_ERROR("qapi_Loc_Get_Single_Shot_V2 failed (err=%d)", error);
	} else {
		osal_status = OSAL_binary_semaphore_take(&semaphore, OSAL_INFINITE_TIME);
		if (OSAL_OK != osal_status) {
			DEBUG_GNSS_LOG_ERROR("OSAL_binary_semaphore_take failed (err=%d)", osal_status);
		}
	}
}

/**
 * @brief Initializes and setup GNSS.
 * 
 */
void LLGNSS_initialize(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);

	qapi_Location_Callbacks_t callbacks = {0};
	qapi_Location_Error_t error = 0;
	OSAL_status_t osal_status = 0;

	if (0 != client_id) { // Already initialized
		return;
	}
	
	callbacks.size = sizeof(qapi_Location_Callbacks_t);
	callbacks.capabilitiesCb = capabilities_callback;
	callbacks.responseCb = response_callback;
	callbacks.trackingCb = tracking_callback;
	callbacks.collectiveResponseCb = collective_response_callback;
	callbacks.singleShotCb = singleshot_callback;

	error = qapi_Loc_Init(&client_id, &callbacks);
	if (QAPI_LOCATION_ERROR_SUCCESS != error) {
		DEBUG_GNSS_LOG_ERROR("qapi_Loc_Init failed (err=%d)", error);
		SNI_throwNativeException(ERROR_GNSS_INITIALIZATION, "Location module initialization failed");
		return;
	}

	add_cleanup_callback(GNSS_cleanup);

	error = qapi_Loc_Set_User_Buffer(client_id, (uint8_t*) userBuffer, (size_t) USER_BUFFER_SIZE);
	if (QAPI_LOCATION_ERROR_SUCCESS != error) {
		DEBUG_GNSS_LOG_ERROR("qapi_Loc_Set_User_Buffer failed (err=%d)", error);

		error = qapi_Loc_Deinit(client_id);
		if (QAPI_LOCATION_ERROR_SUCCESS != error) {
			DEBUG_GNSS_LOG_ERROR("qapi_Loc_Deinit failed (err=%d)", error);
		}
		client_id = 0;

		SNI_throwNativeException(ERROR_GNSS_INITIALIZATION, "Location module buffer assignation failed");
		return;
	}

	osal_status = OSAL_binary_semaphore_create((uint8_t*) "LLGNSS_semaphore", 0, &semaphore);
	if (OSAL_OK != osal_status) {
		SNI_throwNativeException(ERROR_GNSS_INITIALIZATION, "Semaphore creation failed");
		return;
	}
}

/**
 * @brief Starts GNSS positioning.
 * 
 */
void LLGNSS_start(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);

	qapi_Location_Options_t options = {0};
	qapi_Location_Error_t error = 0;

	if (0 != tracking_id) { // Already started
		return;
	}

	options.size = sizeof(qapi_Location_Options_t);
	options.minInterval= 1000; // Time-based tracking (minDistance = 0). The minimum time interval in milliseconds that must elapse between final position reports
	options.minDistance = 0; // Minimum distance in meters that must be traversed between position reports. Setting this interval to 0 will be a pure time-based tracking/batching.

	error = qapi_Loc_Start_Tracking(client_id, &options, &tracking_id);
	if(QAPI_LOCATION_ERROR_SUCCESS != error) {
		DEBUG_GNSS_LOG_ERROR("qapi_Loc_Start_Tracking failed (err=%d)", error);
		SNI_throwNativeException(ERROR_GNSS_TRACKING, NULL);
	}
}

/**
 * @brief Stops GNSS positioning.
 * 
 */
void LLGNSS_stop(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);

	qapi_Location_Error_t error;

	if (0 == tracking_id) { // Not started
		return;
	}

	error = qapi_Loc_Stop_Tracking(client_id, tracking_id);
	if(QAPI_LOCATION_ERROR_SUCCESS == error) {
		tracking_id = 0;
	} else {
		DEBUG_GNSS_LOG_ERROR("qapi_Loc_Stop_Tracking failed (err=%d)", error);
		SNI_throwNativeException(ERROR_GNSS_TRACKING, NULL);
	}
}

/**
 * @brief Closes GNSS.
 * 
 */
void LLGNSS_close(void) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);

	qapi_Location_Error_t error;

	if (0 == client_id) { // Not initialized
		return;
	}

	error = qapi_Loc_Deinit(client_id);
	if (QAPI_LOCATION_ERROR_SUCCESS == error) {
		client_id = 0;
	} else {
		SNI_throwNativeException(ERROR_GNSS_CLOSE, NULL);
		DEBUG_GNSS_LOG_ERROR("qapi_Loc_Deinit failed (err=%d)", error);
	}
}

static void GNSS_cleanup(void* param) {
	DEBUG_GNSS_LOG_INFO("%s()", __func__);

	qapi_Location_Error_t error = 0;
	OSAL_status_t osal_status = 0;

	if (0 == client_id) { // Nothing to clean
		return;
	}

	error = qapi_Loc_Deinit(client_id);
	if (QAPI_LOCATION_ERROR_SUCCESS == error) {
		client_id = 0;
	} else {
		DEBUG_GNSS_LOG_ERROR("qapi_Loc_Deinit failed (err=%d)", error);
	}

	osal_status = OSAL_binary_semaphore_delete(&semaphore);
	if (OSAL_OK != osal_status) {
		DEBUG_GNSS_LOG_ERROR("OSAL_mutex_delete() failed (err=%d)", osal_status);
	}
}
