/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 *
 */

/**
 * @file
 * @brief LLNET_Common implementation over BSD-like API.
 * @author MicroEJ Developer Team
 * @version 1.1.1
 * @date 7 February 2020
 */

#include <stdint.h>
#include <string.h>
#include "qapi_socket.h"
#include "LLNET_Common.h"
#include "async_select.h"
#include "LLNET_CHANNEL_impl.h"
#include "LLNET_CONSTANTS.h"
#include "LLNET_ERRORS.h"
#if LLNET_AF & LLNET_AF_IPV6
#include <ifaddrs.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>
#endif

#ifndef USE_IOCTL_FOR_BLOCKING_OPTION
#include <fcntl.h>
#endif

static int32_t get_socket_timeout(int32_t fd){
	return LLNET_CHANNEL_IMPL_getOption(fd, CPNET_SO_TIMEOUT, 0);
}

int32_t asyncOperation(int32_t fd, SELECT_Operation operation, uint8_t retry){
	// the socket is not readable or writable.
	// A blocking request is added in the dispatch event queue.
	int32_t timeout = get_socket_timeout(fd);
	//check the timeout
	if(timeout < 0){
		//error code
		return timeout;
	}
	if((timeout != 0) && retry){
		//The socket has been configured with a timeout
		//AND there is no available data
		//AND this is the second read (retry==true)
		//THEN timeout has been reached during the first read
		return J_ETIMEDOUT;
	}
	int32_t res = async_select(fd, operation, timeout, NULL);
	if(res == 0){
		// request added in the queue
		return J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
	}
	// requests queue limit reached
	return J_ASYNC_BLOCKING_REQUEST_QUEUE_LIMIT_REACHED;
}


/**
 * @brief Sets the given socket in non-blocking mode or not.
 *
 * @param[in] fd socket file descriptor
 * @param[in] non_blocking true to enable non-blocking, false to enable blocking.
 *
 * @return 0 on success, a negative value on error.
 */
int32_t set_socket_non_blocking(int32_t fd, bool non_blocking){
#ifdef USE_IOCTL_FOR_BLOCKING_OPTION
	int32_t non_blocking_option = non_blocking==true ? 1 : 0;
	int32_t res = llnet_ioctl(fd, FIONBIO, &non_blocking_option);

	return res;
#elif defined(USE_GETSOCKOPT_FOR_BLOCKING_OPTION)
    int32_t value = 0;
    int32_t result = 0;
    if (non_blocking) {
        value = SO_NBIO;
    } else {
        value = SO_BIO;
    }
    return llnet_setsockopt(fd, SOL_SOCKET, SO_NONBLOCK, &value, sizeof(bool));
#else
	int32_t flags = llnet_fcntl(fd, F_GETFL, 0);
	if(flags == -1){
		return -1;
	}

	if(non_blocking == true){
		flags |= O_NONBLOCK;
	}
	else {
		flags &= ~O_NONBLOCK;
	}
	return llnet_fcntl(fd, F_SETFL, flags);
#endif
}

/**
 * @brief Checks if the given socket is in non-blocking mode or not.
 *
 * @param[in] fd socket file descriptor
 *
 * @return true if the socket is non blocking, false if the socket is blocking or an error occurs.
 */
bool is_socket_non_blocking(int32_t fd){
#ifdef USE_IOCTL_FOR_BLOCKING_OPTION
	return false;
#elif defined(USE_GETSOCKOPT_FOR_BLOCKING_OPTION)
    int32_t result = 0;
    int32_t value = 0;
    result = llnet_getsockopt(fd, SOL_SOCKET, SO_NONBLOCK, &value, sizeof(int32_t));
    if (0 > result || SO_BIO == value) {
        return false;
    }
    return true;    
#else
	int32_t flags = llnet_fcntl(fd, F_GETFL, 0);
	if(flags == -1){
		return false;
	}

	return (flags & O_NONBLOCK) != 0;
#endif
}

/**
 * @brief Convert a network error code into a java error code.
 *
 * @param[in] err the error code returned by a BSD-like function.
 *
 * @return an error code defined in LLNET_ERRORS.h.
 */
int32_t map_to_java_exception(int32_t err){
	switch(err){
		case (EBADF):
			return J_EBADF;
		case (ENETUNREACH):
			return J_ENETUNREACH;
		case (EADDRINUSE):
			return J_EADDRINUSE;
		case (EINVAL):
			return J_EINVAL;
		case (ECONNABORTED):
			return J_ECONNABORTED;
		case (ENOPROTOOPT):
			return J_ENOPROTOOPT;
		case (ENOTCONN):
			return J_ENOTCONN;
		case (EAFNOSUPPORT):
			return J_EAFNOSUPPORT;
		case (ECONNREFUSED):
			return J_ECONNREFUSED;
		case (EISCONN):
			return J_EISCONN;
		case (ECONNRESET):
			return J_ECONNRESET;
		case (EMSGSIZE):
			return J_EMSGSIZE;
		case (EPIPE):
			return J_EPIPE;
		case (ETIMEDOUT):
			return J_ETIMEDOUT;
		case (ENOBUFS):
		case (ENOMEM):
			return J_ENOMEM;
		default:
			return J_EUNKNOWN;
	}
}

#if LLNET_AF & LLNET_AF_IPV6

uint32_t getScopeForIp(const char *ip){
	struct ifaddrs *addrs;
	char ipAddress[NI_MAXHOST];
	uint32_t scope = 0;

	// walk over the list of all interface addresses
	getifaddrs(&addrs);
	for (struct ifaddrs *addr = addrs; addr; addr = addr->ifa_next) {
		if (addr->ifa_addr && addr->ifa_addr->sa_family == AF_INET6) { // only interested in ipv6 ones
			getnameinfo(addr->ifa_addr, sizeof(struct sockaddr_in6), ipAddress,
					sizeof(ipAddress), NULL, 0, NI_NUMERICHOST);
			// result actually contains the interface name, so strip it
			for (int i = 0; ipAddress[i]; i++) {
				if (ipAddress[i] == '%') {
					ipAddress[i] = '\0';
					break;
				}
			}
			// if the ip matches, convert the interface name to a scope index
			if (strcmp(ipAddress, ip) == 0) {
				scope = if_nametoindex(addr->ifa_name);
				break;
			}
		}
	}
	// If the scope wasn't set in the name search, set it to LLNET_IPV6_INTERFACE_NAME
	if(0 == scope) {
		scope = if_nametoindex(LLNET_IPV6_INTERFACE_NAME);
	}
	freeifaddrs(addrs);
	return scope;
}

/**
 * @brief  Map the given IPv4 address into an IPv6 address.<p>
 * <code>ipv4_addr</code> and <code>ipv6_addr</code> pointers can reference the same memory address.
 *
 * @param[in] ipv4_addr the IPv4 to map
 * @param[out] ipv6_addr the destination IPv6
 */
void map_ipv4_into_ipv6(in_addr_t* ipv4_addr, struct in6_addr* ipv6_addr){
	// Convert IPv4 into IPv6 directly in the addr array

	if(*(ipv4_addr) == llnet_htonl(INADDR_ANY)){
		// Use IPv6 wildcard address
		*ipv6_addr = in6addr_any;
	}
	else {
		uint8_t* ipv6_addr_u8 = (uint8_t*)ipv6_addr;
		// Destination offset of the IPv4 address: end of struct in6_addr
		int ipv4_dest_offset = sizeof(struct in6_addr) - sizeof(in_addr_t);
		// Copy the IPv4 address at the end of struct in6_addr
		memcpy(ipv6_addr_u8 + ipv4_dest_offset, ipv4_addr, sizeof(in_addr_t));

		// Add 0xFF:0xFF before
		ipv6_addr_u8[ipv4_dest_offset - 1] = 0xFF;
		ipv6_addr_u8[ipv4_dest_offset - 2] = 0xFF;

		// Set to 0 all the other bytes
		memset(ipv6_addr_u8, 0, ipv4_dest_offset - 2);
	}
}

#endif
