/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */

#include <string.h>
#include <LLNET_NETWORKINTERFACE_impl.h>
#include "LLNET_Common.h"
#include "microej_allocator.h"
#include "qapi_dss.h"
#include "qapi_netservices.h"

#ifdef __cplusplus
	extern "C" {
#endif

#define MAX_INTERFACE_NUMBER (4)
#define NETWORK_INTERFACE_ADDRESSES (1)

// ipv4 address info size (tag (1) + IP (4) + prefix (1) + hasBroadcast (1) + broadcast IP (4))
#define IPV4_ADDR_INFO_SIZE 11

// ipv4 address tag
#define IPV4_ADDR_TAG 4

extern qapi_DSS_Hndl_t netctrl_dss_handle;

/**
 * @brief Gets the interface data structure.
 * 
 * @param expected_name The name of the interface.
 * 
 * @param length The length of the interface name.
 * 
 * @return The interface data.
 */
static qapi_Net_Ifnameindex_t get_interface_by_name(uint8_t* expected_name, int32_t length) {
	uint8_t index = 0;
	qapi_Net_Ifnameindex_t interface = {0};
	qapi_Net_Ifnameindex_t interface_list[MAX_INTERFACE_NUMBER];

	qapi_Net_Get_All_Ifnames(&interface_list);

	do {
		interface = interface_list[index++];
	} while(0 != strcmp(expected_name, interface.interface_Name));
	
	return interface;
}

/**
 * @brief Gets the interface data structure.
 * 
 * @param index Index of the interface in the list.
 * 
 * @return The interface data.
 */
static qapi_Net_Ifnameindex_t get_interface_by_index(uint32_t index) {
	qapi_Net_Ifnameindex_t interface = {0};
	qapi_Net_Ifnameindex_t interface_list[MAX_INTERFACE_NUMBER];

	qapi_Net_Get_All_Ifnames(&interface_list);
	
	return interface_list[index];
}

/**
 * Gets the hardware address associated with the given network interface name {@code ifname}.
 * <p>When this method returns, the buffer {@code hwAddr} is filled with the hardware address.
 * @param ifname the network interface name buffer
 * @param ifnameLen the network interface name length
 * @param hwAddr the buffer into which the hardware address will be stored
 * @param hwAddrMaxLength maximum size of hardware address that can be stored in the buffer {@code hwAddr}
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return the hardware address size in bytes, 0 if no hardware address found or a negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_getHardwareAddress(int8_t* ifname, int32_t ifnameLen, int8_t* hwAddr, int32_t hwAddrMaxLength,
		uint8_t retry) {
	uint32_t hardware_address_length = 0;
	int32_t result = 0;

	uint8_t* hardware_address = microej_malloc(hwAddrMaxLength);

	qapi_Net_Ifnameindex_t interface = get_interface_by_name(ifname, ifnameLen);

	result = qapi_Net_Interface_Get_Physical_Address(interface.interface_Name, hardware_address, &hardware_address_length);
	if (0 != result) {
		LLNET_DEBUG_TRACE("%s: Failed to get hardware address, result = %d", __func__, result);
		return J_EUNKNOWN;
	}

	if (hardware_address_length > hwAddrMaxLength) {
		hardware_address_length = hwAddrMaxLength;
	}

	memcpy(hwAddr, &hardware_address, hardware_address_length);
	return hardware_address_length;
}


/**
 * Gets the network interface name at the {@code index} position in the network interfaces list.
 * When this method returns, the {@link ifname} buffer is filled with the network interface name
 * @param index the network interface index
 * @param ifname the buffer into which the network interface name will be stored
 * @param length the buffer length
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return the network interface name length, 0 if no interface found or a negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_getVMInterface(int32_t index, int8_t* ifname, int32_t length, uint8_t retry){
	qapi_Net_Ifnameindex_t interface = get_interface_by_index(index);
	memcpy(ifname, &interface.interface_Name, length);
	return strlen(interface.interface_Name);
}

/**
 * Returns the number of the network interface addresses.
 * @param index the network interface index-position in the interfaces list
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return the number of the network interface addresses, or a negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_getVMInterfaceAddressesCount(int32_t index, uint8_t retry){
	return NETWORK_INTERFACE_ADDRESSES;
}

/**
 * Gets a network interface address in the interface addresses list.
 * <p>When this method returns, the {@code addr} buffer is filled with the network interface address information, using the following layout:
 * <ul> a one-byte tag ({@link IPV4_ADDR_INFO_SIZE} or {@link IPV6_ADDR_INFO_SIZE})
 * <ul> the IP address bytes (4 or 16)
 * <ul> the prefix length in bits (encoded values from 0 to 0x80, meaning 1 to 128)
 * <ul> for an IP v4 address, a one-byte boolean value (0 or 1) telling if there's a broadcast address
 * <ul> for an IP v4 address that has broadcast, the 4 bytes of the broadcast address
 * </ul>
 * @param idxIf the index-position of the interface in the interfaces list
 * @param idxAddr the index-position of the address in the interface addresses list
 * @param addr the buffer into which the address will be stored
 * @param length the buffer length ({@link IPV4_ADDR_TAG} or {@link IPV6_ADDR_TAG})
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return the network interface address size in bytes or a negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_getVMInterfaceAddress(int32_t idxIf, int32_t idxAddr, int8_t* addr, int32_t length, uint8_t retry){
	uint32_t ipv4_address = 0;
	
	qapi_Net_Ifnameindex_t interface = get_interface_by_index(idxIf);
	qapi_Status_t status = qapi_Net_IPv4_Config(interface.interface_Name, QAPI_NET_IPV4CFG_QUERY_E, &ipv4_address, NULL, NULL);
	if (0 != status) {
		LLNET_DEBUG_TRACE("%s: Failed to get IPv4 configuration, result = %d", __func__);
		return J_EUNKNOWN;
	}

	addr[0] = IPV4_ADDR_TAG;
	memcpy(addr[1], &ipv4_address, sizeof(ipv4_address));
	addr[sizeof(ipv4_address) + 1] = 1; // No information on prefix.
	addr[sizeof(ipv4_address) + 2] = 0;

	return IPV4_ADDR_INFO_SIZE;
}

/**
 * Returns the number of network interfaces.
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return the number of network interfaces or a negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_getVMInterfacesCount(uint8_t retry){
	qapi_Net_Ifnameindex_t ifnames[MAX_INTERFACE_NUMBER];
	return qapi_Net_Get_All_Ifnames(&ifnames);
}

/**
 * Checks if the given network interface {@code ifname} is a loopback interface.
 * @param ifname the network interface name
 * @param length the network interface name length
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return 0 if true, 1 if false or negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_isLoopback(int8_t* ifname, int32_t length, uint8_t retry){
	return 1; // Information not available.
}

/**
 * Checks if the given network interface {@code ifname} is a point-to-point interface.
 * @param ifname the network interface name
 * @param length the network interface name length
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return 0 if true, 1 if false or negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_isPointToPoint(int8_t* ifname, int32_t length, uint8_t retry){
	return 1; // Information not available.
}

/**
 * Checks if the given network interface {@code ifname} is up.
 * @param ifname the network interface name
 * @param length the network interface name length
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return 0 if true, 1 if false or negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_isUp(int8_t* ifname, int32_t length, uint8_t retry){
	qapi_Net_Ifnameindex_t interface = get_interface_by_name(ifname, length);
	return interface.if_Is_Up;
}

/**
 * Checks if the given network interface {@code ifname} supports multicast.
 * @param name the network interface name
 * @param length the network interface name length
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return 0 if true, 1 if false or negative error code
 * @see {@link LLNET_ERRORS} header file for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_supportsMulticast(int8_t* ifname, int32_t length, uint8_t retry){
	return 1; // Information not available.
}

/**
 * @param name the network interface name
 * @param length the network interface name length
 * @param retry true when the previous call returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
 * and the calling process repeats the call to this operation for its completion
 * @return the MTU as an int, or a negative error code
 * @see {@link NetErrors} for error codes
 */
int32_t LLNET_NETWORKINTERFACE_IMPL_getMTU(int8_t* ifname, int32_t length, uint8_t retry){
	int32_t mtu = 0;
	
	qapi_Status_t status = qapi_DSS_Get_Link_Mtu(netctrl_dss_handle, &mtu);
	if (QAPI_OK != status) {
		LLNET_DEBUG_TRACE("%s: Failed to get MTU, status = %d", status);
		return J_EUNKNOWN;
	}

	return mtu;
}

#ifdef __cplusplus
	}
#endif
