/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */

/**
 * @file
 * @brief SSL low level implementation : SSL context
 * @author MicroEJ Developer Team
 * @version 1.0.0
 * @date 16 September 2020
 */

#include <LLNET_SSL_CONTEXT_impl.h>
#include "LLNET_SSL_configuration.h"
#include "qapi_status.h"
#include "qapi_ssl.h"
#include "microej_allocator.h"

#ifdef __cplusplus
	extern "C" {
#endif

/**
 * Refresh the SSL context object to keep truststore and keystore aligned with the given context.
 * @param contextID the SSL context ID.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
static int32_t updateCertificates(int32_t contextID);

/**
 * Creates a new client SSL Context based on the given SSL/TLS protocol version.
 * @param protocol the SSL/TLS protocol version (0 for SSLv3, 1 for TLSv1, 2 for TLSv1.1 and 3 for TLSv1.2).
 * @param isClientContext controls whether the underlying context will be created to manage ssl client or ssl server sockets. 
 * <code>true</code> for client sockets; <code>false</code> for server sockets.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call.
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return the new SSL Context ID on success, {@link J_CREATE_SSL_CONTEXT_ERROR} on error (for example if not enough space to create the context).
 */
int32_t LLNET_SSL_CONTEXT_IMPL_createContext(int32_t protocol, uint8_t isClientContext, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(protocol=%d, isClientContext=%d, retry=%d)", __func__, (int)protocol, isClientContext, retry);

	LLNET_SSL_context_t* context;
	qapi_Net_SSL_Obj_Hdl_t handle;

	if (false == isClientContext) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Server role not supported");
		return J_CREATE_SSL_CONTEXT_ERROR;
	}

	context = (LLNET_SSL_context_t*) microej_malloc(sizeof(LLNET_SSL_context_t));
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] SSL context allocation failed");
		return J_CREATE_SSL_CONTEXT_ERROR;
	}

	handle = qapi_Net_SSL_Obj_New(QAPI_NET_SSL_CLIENT_E);
	if (QAPI_NET_SSL_INVALID_HANDLE == handle) {
		microej_free(context);
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Obj_New failed");
		return J_CREATE_SSL_CONTEXT_ERROR;
	}

	context->handle = handle;
	context->protocol = protocol;
	context->certificate_buffer_len = 0;
	context->cert_chain = NULL;

	LLNET_SSL_DEBUG_TRACE("%s(protocol=%d, isClientContext=%d, retry=%d) context=%d", __func__, (int)protocol, isClientContext, retry, (int)context);
	return (int32_t) context;
}

/**
 * Releases the SSL Context identified by the given contextID.
 * @param contextID the SSL context ID.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
int32_t LLNET_SSL_CONTEXT_IMPL_closeContext(int32_t contextID, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d)\n", __func__, (int)contextID, retry);

	LLNET_SSL_context_t* context;
	qapi_Status_t status;
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}

	status = qapi_Net_SSL_Obj_Free(context->handle);
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Obj_Free failed");
		return J_UNKNOWN_ERROR;
	}

	if (NULL != context->cert_chain) {
		microej_free(context->cert_chain);
	}
	microej_free(context);

	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d) ret=0\n", __func__, (int)contextID, retry);
	return J_SSL_NO_ERROR;
}

/**
 * Adds the given certificate as trusted certificate into the given SSL context.
 * @param contextID the SSL Context ID.
 * @param cert the certificate buffer.
 * @param off the offset in the buffer at which the certificate content started.
 * @param len the certificate content length.
 * @param format the certificate encoded format type (0 for PEM format and 1 for DER format).
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
int32_t LLNET_SSL_CONTEXT_IMPL_addTrustedCert(int32_t contextID, uint8_t *cert, int32_t off, int32_t len, int32_t format, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d)", __func__, (int)contextID, retry);

	LLNET_SSL_context_t* context;
	qapi_Net_SSL_Cert_Info_t certificate;
	qapi_NET_SSL_CA_Info_t ca_info;
	qapi_Status_t status;
	int32_t ret;
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}

	certificate.cert_Type = QAPI_NET_SSL_CA_LIST_E;
	certificate.info.ca_List.ca_Cnt = 1;

	ca_info.ca_Buf = cert + off;
	ca_info.ca_Size = len;
	certificate.info.ca_List.ca_Info[0] = &ca_info;

	status = qapi_Net_SSL_Cert_Convert_And_Store(&certificate, "ca_contextID"); // TODO add the context id to the name of the certificate
	if (status < 0) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_Convert_And_Store failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	ret = updateCertificates(contextID);
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d) ret=%d", __func__, (int)contextID, retry, (int)ret);
	return ret;
}

/**
 * Removes all trusted certificates from the given SSL context.
 * @param contextID the SSL Context ID.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
int32_t LLNET_SSL_CONTEXT_IMPL_clearTrustStore(int32_t contextID, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d)", __func__, (int)contextID, retry);

	qapi_Status_t status;
	qapi_Net_SSL_Cert_List_t list;
	int32_t ret;
	
	status = qapi_Net_SSL_Cert_List(QAPI_NET_SSL_CA_LIST_E, &list);
	if (status < 0) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_List failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}
	
	for (int i = 0; i < status; i++) {
		status = qapi_Net_SSL_Cert_delete(list.name[i], QAPI_NET_SSL_CA_LIST_E);
		if (status < 0) {
			LLNET_SSL_DEBUG_TRACE("[WARNING] Could not delete the CA certificate \"%s\"", list.name[i]);
		}
	}

	ret = updateCertificates(contextID);
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d) ret=%d", __func__, (int)contextID, retry, ret);
	return ret;
}

/**
 * Removes all credentials (private key and associated certificates) from the given ssl context.
 * @param contextID the ssl context ID.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
int32_t LLNET_SSL_CONTEXT_IMPL_clearKeyStore(int32_t contextID, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d)", __func__, (int)contextID, retry);

	qapi_Status_t status;
	qapi_Net_SSL_Cert_List_t list;
	int32_t ret;
	
	status = qapi_Net_SSL_Cert_List(QAPI_NET_SSL_CERTIFICATE_E, &list);
	if (status < 0) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_List failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}
	
	for (int i = 0; i < status; i++) {
		status = qapi_Net_SSL_Cert_delete(list.name[i], QAPI_NET_SSL_CERTIFICATE_E);
		if (status < 0) {
			LLNET_SSL_DEBUG_TRACE("[WARNING] Could not delete the key and associated certificate \"%s\"", list.name[i]);
		}
	}

	ret = updateCertificates(contextID);
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, retry=%d) ret=%d", __func__, (int)contextID, retry, ret);
	return ret;
}

/**
 * Sets the certificate of the context private key.
 * @param contextID the ssl context ID.
 * @param cert the certificate of the context private key.
 * @param offset the offset in the buffer at which the certificate content started.
 * @param len the certificate content length.
 * @param format the certificate encoded format type (0 for PEM format and 1 for DER format).
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 *
 */
int32_t LLNET_SSL_CONTEXT_IMPL_setCertificate(int32_t contextID, uint8_t* cert, int32_t offset, int32_t len, int32_t format, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, format=%d, retry=%d)", __func__, (int)contextID, (int)len, (int)format, retry);

	LLNET_SSL_context_t* context;
	qapi_Net_SSL_Cert_Info_t certificate;
	qapi_Status_t status;
	int32_t ret;
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}

	if (0 == context->certificate_buffer_len) {
		if (LLNET_SSL_MAX_CERTIFICATE_SIZE < len) {
			LLNET_SSL_DEBUG_TRACE("[ERROR] Maximum key size exceeded (actual=%d, maximum=%d)", len, LLNET_SSL_MAX_CERTIFICATE_SIZE);
			return J_UNKNOWN_ERROR;
		}

		memcpy(context->certificate_buffer, cert + offset, len);
		context->certificate_buffer_len = len;

		LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, format=%d, retry=%d) ret=0", __func__, (int)contextID, (int)len, (int)format, retry);
		return J_SSL_NO_ERROR;
	}

	certificate.cert_Type = QAPI_NET_SSL_CERTIFICATE_E;
	certificate.info.cert.cert_Buf = cert + offset;
	certificate.info.cert.cert_Size = len;
	certificate.info.cert.key_Buf = context->certificate_buffer;
	certificate.info.cert.key_Size = context->certificate_buffer_len;
	certificate.info.cert.pass_Key = context->password;

	status = qapi_Net_SSL_Cert_Convert_And_Store(&certificate, "certificate_contextID"); // TODO add the context id to the name of the certificate
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_Convert_And_Store failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	ret = updateCertificates(contextID);
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, format=%d, retry=%d) ret=%d", __func__, (int)contextID, (int)len, (int)format, retry, ret);
	return ret;
}

/**
 * Sets the context private key.
 * @param contextID the ssl context ID.
 * @param privateKey the context private key.
 * @param offset the offset in the buffer at which the private key content started.
 * @param len the private key content length.
 * @param password the password to recover the private key if it is encrypted.
 * @param passwordOffset the offset in the password buffer at which the password content started.
 * @param passwordLen the password length.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
int32_t LLNET_SSL_CONTEXT_IMPL_setPrivateKey(int32_t contextID, uint8_t* privateKey, int32_t offset, int32_t len, uint8_t* password,
		int32_t passwordOffset, int32_t passwordLen, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, passwordLen=%d, retry=%d)", __func__, (int)contextID, (int)len, (int)passwordLen, retry);

	LLNET_SSL_context_t* context;
	qapi_Net_SSL_Cert_Info_t certificate;
	qapi_Status_t status;
	int32_t ret;
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}

	if (0 == context->certificate_buffer_len) {
		if (LLNET_SSL_MAX_CERTIFICATE_SIZE < len) {
			LLNET_SSL_DEBUG_TRACE("[ERROR] Maximum key size exceeded (actual=%d, maximum=%d)", len, LLNET_SSL_MAX_CERTIFICATE_SIZE);
			return J_UNKNOWN_ERROR;
		}
		if (LLNET_SSL_MAX_PASSWORD_SIZE - 1 < passwordLen) {
			LLNET_SSL_DEBUG_TRACE("[ERROR] Maximum password size exceeded (actual=%d, maximum=%d)", passwordLen, LLNET_SSL_MAX_PASSWORD_SIZE - 1);
			return J_UNKNOWN_ERROR;
		}

		memcpy(context->certificate_buffer, privateKey + offset, len);
		context->certificate_buffer_len = len;
		memcpy(context->password, password + passwordOffset, passwordLen);
		
		LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, passwordLen=%d, retry=%d) ret=0", __func__, (int)contextID, (int)len, (int)passwordLen, retry);
		return J_SSL_NO_ERROR;
	}

	certificate.cert_Type = QAPI_NET_SSL_CERTIFICATE_E;
	certificate.info.cert.cert_Buf = context->certificate_buffer;
	certificate.info.cert.cert_Size = context->certificate_buffer_len;
	certificate.info.cert.key_Buf = privateKey + offset;
	certificate.info.cert.key_Size = len;
	memcpy(context->password, password + passwordOffset, passwordLen);
	context->password[passwordLen] = '\0';
	certificate.info.cert.pass_Key = context->password;

	status = qapi_Net_SSL_Cert_Convert_And_Store(&certificate, "certificate_contextID"); // TODO add the context id to the name of the certificate
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_Convert_And_Store failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	ret = updateCertificates(contextID);
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, passwordLen=%d, retry=%d) ret=%d", __func__, (int)contextID, (int)len, (int)passwordLen, retry, ret);
	return ret;
}

/**
 * Initializes the chain buffer and returns its total size.
 * @param contextID the ssl context ID.
 * @param nbChainCerts the number of certificates in the chain.
 * @param chainCertsTotalSize the total size of certificates in the chain.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return the total size of the chain buffer on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
int32_t LLNET_SSL_CONTEXT_IMPL_initChainBuffer(int32_t contextID, int32_t nbChainCerts, int32_t chainCertsTotalSize, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, nbChainCerts=%d, chainCertsTotalSize=%d, retry=%d)", __func__, (int)contextID, (int)nbChainCerts, (int)chainCertsTotalSize, retry);

	LLNET_SSL_context_t* context;

	if (0 > chainCertsTotalSize) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid parameter (chainCertsTotalSize=%d)", chainCertsTotalSize);
		return J_UNKNOWN_ERROR;
	}
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}

	context->cert_chain_info.cert_Type = QAPI_NET_SSL_CA_LIST_E;
	context->cert_chain_info.info.ca_List.ca_Cnt = nbChainCerts;
	context->cert_in_chain = 0;
	context->cert_chain = microej_malloc(chainCertsTotalSize);
	context->cert_chain_offset = 0;

	LLNET_SSL_DEBUG_TRACE("%s(context=%d, nbChainCerts=%d, chainCertsTotalSize=%d, retry=%d) ret=%d", __func__, (int)contextID, (int)nbChainCerts, (int)chainCertsTotalSize, retry, (int)chainCertsTotalSize);
	return chainCertsTotalSize;
}

/**
 * Adds the given certificate to the certificate chain buffer associated with the context private key.
 * @param contextID the ssl context ID.
 * @param cert the certificate of the context private key.
 * @param offset the offset in the buffer at which the certificate content started
 * @param len the certificate content length
 * @param format the certificate encoded format type (0 for PEM format and 1 for DER format).
 * @param chainBufferSize the chain buffer size.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
int32_t LLNET_SSL_CONTEXT_IMPL_addChainCertificate(int32_t contextID, uint8_t* cert, int32_t offset, int32_t len, int32_t format, int32_t chainBufferSize, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, format=%d, chainBufferSize=%d, retry=%d)", __func__, (int)contextID, (int)len, (int)format, (int)chainBufferSize, retry);

	LLNET_SSL_context_t* context;
	qapi_NET_SSL_CA_Info_t* ca_info;
	qapi_Status_t status;
	int32_t ret;
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}

	if (context->cert_chain_offset + len > chainBufferSize) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Not enough space in the certificate chain buffer (requested=%d, available=%d)", len, chainBufferSize - context->cert_chain_offset);
		return J_UNKNOWN_ERROR;
	}

	ca_info = microej_malloc(sizeof(qapi_NET_SSL_CA_Info_t));
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Certificate allocation failed");
		return J_CREATE_SSL_CONTEXT_ERROR;
	}

	memcpy(context->cert_chain + context->cert_chain_offset, cert + offset, len);
	ca_info->ca_Buf = context->cert_chain + context->cert_chain_offset;
	ca_info->ca_Size = len;
	context->cert_chain_offset += len;
	context->cert_chain_info.info.ca_List.ca_Info[context->cert_in_chain] = ca_info;
	context->cert_in_chain++;

	status = qapi_Net_SSL_Cert_Convert_And_Store(&context->cert_chain_info, "ca_contextID"); // TODO add the context id to the name of the certificate
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_Convert_And_Store failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	ret = updateCertificates(contextID);
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, len=%d, format=%d, chainBufferSize=%d, retry=%d) ret=%d", __func__, (int)contextID, (int)len, (int)format, (int)chainBufferSize, retry, (int)ret);
	return ret;
}

/**
 * Refresh the SSL context object to keep truststore and keystore aligned with the given context.
 * @param contextID the SSL context ID.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code.
 * @see {@link LLNET_SSL_ERRORS} header file for error codes.
 */
static int32_t updateCertificates(int32_t contextID) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d)", __func__, contextID);

	LLNET_SSL_context_t* context;
    qapi_Net_SSL_Cert_List_t list;
    qapi_Status_t status;
	qapi_Net_SSL_Obj_Hdl_t handle;
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}
	
	status = qapi_Net_SSL_Obj_Free(context->handle);
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Obj_Free failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}
	
	handle = qapi_Net_SSL_Obj_New(QAPI_NET_SSL_CLIENT_E);
	if (QAPI_NET_SSL_INVALID_HANDLE == handle) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Obj_New failed");
		return J_CREATE_SSL_CONTEXT_ERROR;
	}
	context->handle = handle;

    memset(&list, 0, sizeof(qapi_Net_SSL_Cert_List_t));
    status = qapi_Net_SSL_Cert_List(QAPI_NET_SSL_CA_LIST_E, &list);
	if (0 > status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_List failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	for(int i = 0; i < status; i++) {
		if (0 == strncmp("ca_contextID", list.name[i], 64)) { // FIXME
			status = qapi_Net_SSL_Cert_Load(context->handle, QAPI_NET_SSL_CA_LIST_E, list.name[i]);
			if (0 > status) {
				LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_Load(handle=%d, QAPI_NET_SSL_CA_LIST_E, name=%s) failed (err=%d)", context->handle, list.name[i], status);
				return J_UNKNOWN_ERROR;
			}
		}
	}

    memset(&list, 0, sizeof(qapi_Net_SSL_Cert_List_t));
    status = qapi_Net_SSL_Cert_List(QAPI_NET_SSL_CERTIFICATE_E, &list);
	if (0 > status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_List failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	for(int i = 0; i < status; i++) {
		if (0 == strncmp("certificate_contextID", list.name[i], 64)) { // FIXME
			status = qapi_Net_SSL_Cert_Load(context->handle, QAPI_NET_SSL_CERTIFICATE_E, list.name[i]);
			if (0 > status) {
				LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Cert_Load failed (err=%d)", status);
				return J_UNKNOWN_ERROR;
			}
		}
	}

	LLNET_SSL_DEBUG_TRACE("%s(context=%d) ret=0", __func__, contextID);
	return J_SSL_NO_ERROR;
}

#ifdef __cplusplus
	}
#endif
