/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */

/**
 * @file
 * @brief SSL low level implementation : SSL socket
 * @author MicroEJ Developer Team
 * @version 1.0.0
 * @date 16 September 2020
 */

#include <LLNET_SSL_SOCKET_impl.h>
#include "LLNET_SSL_configuration.h"
#include "qapi_status.h"
#include "qapi_ssl.h"
#include "LLNET_configuration.h"
#include "LLNET_ERRORS.h"
#include "microej_allocator.h"
#include "LLNET_Common.h"

#ifdef __cplusplus
	extern "C" {
#endif

static int32_t asyncOperationWrapper(int32_t fd, SELECT_Operation operation, uint8_t retry);

/**
 * Initializes SSL/TLS stack components
 * @return {@link J_SSL_NO_ERROR} on success
 */
int32_t LLNET_SSL_SOCKET_IMPL_initialize(void) {
	LLNET_SSL_DEBUG_TRACE("%s()", __func__);
	LLNET_SSL_DEBUG_TRACE("%s() ret=0", __func__);
	return J_SSL_NO_ERROR;
}

/**
 * Creates a new SSL Socket using the given SSL Context and the given underlying
 * Socket file descriptor {@code fd} for I/O operations
 * @param contextID the SSL Context ID
 * @param fd the underlying socket file descriptor used for I/O operations
 * @param hostname the server hostname
 * @param hostnameLen the length of the server hostname
 * @param autoclose true if the underlying socket will be closed when this SSL Socket is closed; false otherwise
 * @param useClientMode configures the socket to use client (or server) mode when handshaking. <code>true</code> if the socket should start its
 * handshaking in "client" mode; <code>false</code> if the socket should start its handshaking in "server" mode
 * @param needClientAuth configures the socket to require or not the client authentication. This option is only useful for sockets in the
 * server mode.<code>true</code> if the client authentication is required; <code>false</code> otherwise.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return the new SSL Socket ID on success, {@link J_CREATE_SSL_SOCKET_ERROR} on error
 */
int32_t LLNET_SSL_SOCKET_IMPL_create(int32_t contextID, int32_t fd, uint8_t* hostname, int32_t hostnameLen, uint8_t autoclose, uint8_t useClientMode, uint8_t needClientAuth, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(context=%d, fd=%d, autoclose=%d, useClientMode=%d, needClientAuth=%d, retry=%d)", __func__, (int)contextID, (int)fd, autoclose, useClientMode, needClientAuth, retry);

	LLNET_SSL_context_t* context;
	qapi_Net_SSL_Config_t  config;
	qapi_Status_t status;
	qapi_Net_SSL_Con_Hdl_t connection;
	LLNET_SSL_socket_t* socket;
	
	context = (LLNET_SSL_context_t*) contextID;
	if (NULL == context) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL context");
		return J_UNKNOWN_ERROR;
	}

	/* 1. create SSL connection handler for an SSL object
	 *    (need be executed after load certificate/psk)
	 */
	connection = qapi_Net_SSL_Con_New(context->handle, QAPI_NET_SSL_TLS_E);
	if (QAPI_NET_SSL_INVALID_HANDLE == connection) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Con_New failed");
		return J_CREATE_SSL_SOCKET_ERROR;
	}

	/* 2. Configures an SSL connection regarding protocol and cipher,
	 *    certificate verify criteria...
	 */
	memset(&config, 0x00, sizeof(qapi_Net_SSL_Config_t));
	switch (context->protocol) {
	case 0:
		LLNET_SSL_DEBUG_TRACE("[ERROR] SSLv3 not supported");
		return J_CREATE_SSL_SOCKET_ERROR;
	case 1:
		config.protocol = QAPI_NET_SSL_PROTOCOL_TLS_1_0;
		break;
	case 2:
		config.protocol = QAPI_NET_SSL_PROTOCOL_TLS_1_1;
		break;
	case 3:
		config.protocol = QAPI_NET_SSL_PROTOCOL_TLS_1_2;
		break;
	default:
		LLNET_SSL_DEBUG_TRACE("[ERROR] Unknown SSL protocol");
		return J_CREATE_SSL_SOCKET_ERROR;
	}
	config.cipher[0] = QAPI_NET_TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256;
	config.cipher[1] = QAPI_NET_TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256;
	config.cipher[2] = QAPI_NET_TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256;
	config.cipher[3] = QAPI_NET_TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256;
	config.cipher[4] = QAPI_NET_TLS_DHE_RSA_WITH_AES_128_CBC_SHA256;
	config.cipher[5] = QAPI_NET_TLS_DHE_RSA_WITH_AES_256_CBC_SHA256;
	config.cipher[6] = QAPI_NET_TLS_DHE_RSA_WITH_AES_128_GCM_SHA256;
	config.cipher[7] = QAPI_NET_TLS_RSA_WITH_AES_128_GCM_SHA256;
	config.verify.domain = true;
	config.verify.time_Validity = true;
	config.verify.send_Alert = true;
	strncpy(config.verify.match_Name, (char *) hostname, (hostnameLen > QAPI_NET_SSL_MAX_CERT_NAME_LEN) ? QAPI_NET_SSL_MAX_CERT_NAME_LEN : hostnameLen);
	config.max_Frag_Len = 16384;
	config.max_Frag_Len_Neg_Disable = 1;
	config.sni_Name = (char*)hostname;
	config.sni_Name_Size = hostnameLen > 0 ? hostnameLen - 1 : 0; // hostname is null-terminated

  	status = qapi_Net_SSL_Configure(connection, &config);
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Configure failed (err=%d)", status);
		return J_CREATE_SSL_SOCKET_ERROR;
	}

	status = qapi_Net_SSL_Fd_Set(connection, fd);
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Fd_Set failed (err=%d)", status);
		return J_CREATE_SSL_SOCKET_ERROR;
	}
	
	socket = (LLNET_SSL_socket_t*) microej_malloc(sizeof(LLNET_SSL_socket_t));
	if (NULL == socket) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] SSL socket allocation failed");
		return J_CREATE_SSL_SOCKET_ERROR;
	}

	socket->connection = connection;
	socket->fd = fd;
	socket->available = 0;

	LLNET_SSL_DEBUG_TRACE("%s(context=%d, fd=%d, autoclose=%d, useClientMode=%d, needClientAuth=%d, retry=%d) ret=%d", __func__, (int)contextID, (int)fd, autoclose, useClientMode, needClientAuth, retry, (int)socket);
	return (int32_t) socket;
}

/**
 * Sends close_notify alert and reads the received close_notify to clear the socket input stream if autoclose is set to false
 * and finally releases the SSL Socket identified by the given ssl ID.
 * @param sslID the SSL Socket ID
 * @param fd the associated underlying socket file descriptor
 * @param autoclose if true, no need to clear the input stream, the underlying socket will be closed;
 * otherwise read the received close_notify alert to clear the socket input stream.
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code
 * @see {@link LLNET_SSL_ERRORS} header file for error codes
 */
int32_t LLNET_SSL_SOCKET_IMPL_close(int32_t sslID, int32_t fd, uint8_t autoclose, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, autoclose=%d, retry=%d)\n", __func__, (int)sslID, (int)fd, autoclose, retry);

	LLNET_SSL_socket_t* socket;
	qapi_Net_SSL_Con_Hdl_t connection;
	qapi_Status_t status;

	socket = (LLNET_SSL_socket_t*) sslID;
	if (NULL == socket) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL socket");
		return J_UNKNOWN_ERROR;
	}
	connection = socket->connection;
	
	status = qapi_Net_SSL_Shutdown(connection);
	if (QAPI_OK != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Shutdown failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, autoclose=%d, retry=%d) ret=0\n", __func__, (int)sslID, (int)fd, autoclose, retry);
	return J_SSL_NO_ERROR;
}

/**
 * Releases the SSL Socket identified by the given ssl ID.
 * @param sslID the SSL Socket ID
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code
 * @see {@link LLNET_SSL_ERRORS} header file for error codes
 */
int32_t LLNET_SSL_SOCKET_IMPL_freeSSL(int32_t sslID, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, retry=%d)\n", __func__, (int)sslID, retry);

	microej_free((LLNET_SSL_socket_t*) sslID);

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, retry=%d) ret=0\n", __func__, (int)sslID, retry);
	return J_SSL_NO_ERROR;
}

/**
 * Initiates the client handshaking
 * @param sslID the SSL Socket ID
 * @param fd the associated underlying socket file descriptor
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code
 * @see {@link LLNET_SSL_ERRORS} header file for error codes
 */
int32_t LLNET_SSL_SOCKET_IMPL_initialClientHandShake(int32_t sslID, int32_t fd, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, retry=%d)", __func__, (int)sslID, (int)fd, retry);

	LLNET_SSL_socket_t* socket;
	qapi_Net_SSL_Con_Hdl_t connection;
	qapi_Status_t status;

	socket = (LLNET_SSL_socket_t*) sslID;
	if (NULL == socket) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL socket");
		return J_UNKNOWN_ERROR;
	}
	connection = socket->connection;

	status = qapi_Net_SSL_Connect(connection);
	if (QAPI_SSL_OK_HS != status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Connect failed (err=%d)", status);
		return J_UNKNOWN_ERROR; // FIXME
	}

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, retry=%d) ret=0", __func__, (int)sslID, (int)fd, retry);
	return J_SSL_NO_ERROR;
}

/**
 * Initiates the server handshaking
 * @param sslID the SSL Socket ID
 * @param fd the associated underlying socket file descriptor
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return {@link J_SSL_NO_ERROR} on success or a negative error code
 * @see {@link LLNET_SSL_ERRORS} header file for error codes
 */
int32_t LLNET_SSL_SOCKET_IMPL_initialServerHandShake(int32_t sslID, int32_t fd, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, retry=%d)", __func__, (int)sslID, (int)fd, retry);
	SNI_throwNativeIOException(-1, ": Server role is not supported.");
	return J_UNKNOWN_ERROR; // Server role is not supported
}

/**
 * Reads len bytes from the SSL Socket internal read buffer into the specified buffer {@code buf}
 * @param sslID the SSL Socket ID
 * @param fd the associated underlying socket file descriptor
 * @param buf the buffer into which the data is read
 * @param off the start offset in the buffer at which the data is written
 * @param len the maximum number of data to read
 * @param retry true if the calling process repeats the call to this operation for its completion because the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return the number of bytes read on success; {@link J_EOF} if there is no more data when the end of the stream
 * has been reached or a negative error code
 * @see {@link LLNET_SSL_ERRORS} header file for error codes
 */
int32_t LLNET_SSL_SOCKET_IMPL_read(int32_t sslID, int32_t fd, int8_t* buf, int32_t off, int32_t len, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d)", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry);

	LLNET_SSL_socket_t* socket;
	qapi_Net_SSL_Con_Hdl_t connection;
	qapi_Status_t status;
	fd_set fdset;
	int32_t read, ret;
	
	socket = (LLNET_SSL_socket_t*) sslID;
	if (NULL == socket) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL socket");
		return J_UNKNOWN_ERROR;
	}
	connection = socket->connection;
	read = socket->available;

	if (read > 0) {
		if (read >= len) {
			memcpy(buf + off, socket->available_buffer, len);
			memcpy(socket->available_buffer, socket->available_buffer + len, read - len);
			socket->available -= len;
			LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) ret=%d", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry, (int)status, len);
			return len;
		}
		
		memcpy(buf + off, socket->available_buffer, read);
		off += read;
		len -= read;
		socket->available = 0;
	}
	
	qapi_fd_zero(&fdset);
	qapi_fd_set(fd, &fdset);
	if (qapi_select(&fdset, NULL, NULL, 0) == 0) {
		ret = asyncOperationWrapper(fd, SELECT_READ, retry);
		LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) ret=%d", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry, (int)ret);
		return ret;
	}

	status = qapi_Net_SSL_Read(connection, buf + off, len);
	if (0 > status) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) ret=-1", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry, (int)status);
		return J_EOF;
	} else {
		read += status; // number of bytes read
	}

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) ret=%d", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry, (int)status, read);
	return read;
}

/**
 * Returns the number of bytes available in the SSL Socket internal read buffer
 * @param sslID the SSL Socket ID
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return the number of available bytes on success or a negative error code
 * @see {@link LLNET_SSL_ERRORS} header file for error codes
 */
int32_t LLNET_SSL_SOCKET_IMPL_available(int32_t sslID, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, retry=%d)", __func__, (int)sslID, retry);

	LLNET_SSL_socket_t* socket;
	qapi_Net_SSL_Con_Hdl_t connection;
	int32_t fd = socket->fd;
	fd_set fdset;
	qapi_Status_t status;

	socket = (LLNET_SSL_socket_t*) sslID;
	if (NULL == socket) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL socket");
		return J_UNKNOWN_ERROR;
	}
	connection = socket->connection;
	
	qapi_fd_zero(&fdset);
	qapi_fd_set(fd, &fdset);
	if (0 < qapi_select(&fdset, NULL, NULL, 0)) {
		status = qapi_Net_SSL_Read(connection, socket->available_buffer + socket->available, SSL_AVAILABLE_BUFFER_SIZE - socket->available);
		if (0 > status) {
			LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Read failed (err=%d)", status);
			return J_UNKNOWN_ERROR;
		}

		socket->available += status; // number of bytes read
	}

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, retry=%d) ret=%d", __func__, (int)sslID, retry, socket->available);
	return socket->available; // number of bytes read
}

/**
 * Writes len bytes from the specified buffer {@code buf} starting at the offset to the SSL Socket output stream
 * @param sslID the SSL Socket ID
 * @param fd the associated underlying socket file descriptor
 * @param buf the buffer of data to write
 * @param off the offset in the data buffer
 * @param len the number of data to write
 * @param retry true if the calling process repeats the call to this operation for its completion when the previous call
 * has returned {@link J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} to indicate that the operation was not completed.
 * @return the number of bytes written on success or a negative error code
 * @see {@link LLNET_SSL_ERRORS} header file for error codes
 */
int32_t LLNET_SSL_SOCKET_IMPL_write(int32_t sslID, int32_t fd, int8_t* buf, int32_t off, int32_t len, uint8_t retry) {
	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d)", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry);

	LLNET_SSL_socket_t* socket;
	qapi_Net_SSL_Con_Hdl_t connection;
	qapi_Status_t status;
	fd_set fdset;
	int32_t ret;

	socket = (LLNET_SSL_socket_t*) sslID;
	if (NULL == socket) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] Invalid SSL socket");
		return J_UNKNOWN_ERROR;
	}
	connection = socket->connection;

	qapi_fd_zero(&fdset);
	qapi_fd_set(fd, &fdset);
	if (qapi_select(NULL, &fdset, NULL, 0) == 0) {
		ret = asyncOperationWrapper(fd, SELECT_WRITE, retry);
		LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) ret=%d", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry, (int)ret);
		return ret;
	}

	status = qapi_Net_SSL_Write(connection, buf + off, len);
	if (0 > status) {
		LLNET_SSL_DEBUG_TRACE("[ERROR] qapi_Net_SSL_Write failed (err=%d)", status);
		return J_UNKNOWN_ERROR;
	}

	LLNET_SSL_DEBUG_TRACE("%s(ssl=%d, fd=%d, offset=%d, length=%d, retry=%d) ret=%d", __func__, (int)sslID, (int)fd, (int)off, (int)len, retry, status);
	return status; // number of bytes written
}

static int32_t asyncOperationWrapper(int32_t fd, SELECT_Operation operation, uint8_t retry){
	int32_t ret = asyncOperation(fd, operation, retry);

	if (ret == J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
		return J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
	} else if (ret == J_ASYNC_BLOCKING_REQUEST_QUEUE_LIMIT_REACHED){
		return J_BLOCKING_QUEUE_LIMIT_REACHED;
	} else if (ret == J_ETIMEDOUT){
		return J_SOCKET_TIMEOUT;
	} else if (ret == J_EUNKNOWN){
		return J_UNKNOWN_ERROR;
	} else {
		return ret;
	}
}


#ifdef __cplusplus
	}
#endif
