/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */

#include <microej_allocator.h>
#include "gina.h"

static TX_BYTE_POOL *byte_pool;
static int32_t module_pool[MODULE_POOL_SIZE/4];

/**
 * @brief Initializes the allocator.
 * @return 0 on success. A negative error code otherwise.
 */
int32_t microej_allocator_initialize() {
	int status;

	status = txm_module_object_allocate(&byte_pool, sizeof(TX_BYTE_POOL));
	if (TX_SUCCESS != status) {
		return status;
	}

	status = tx_byte_pool_create(byte_pool, "module pool", module_pool, MODULE_POOL_SIZE);
	if (TX_SUCCESS != status) {
		return status;
	}

	return 0; // Success
}

/**
 * @brief Allocate a memory area and return the pointer to the allocated memory.
 * @param[in] size Number of bytes to allocate.
 * @return A pointer to the allocated memory. NULL if allocation failed.
 */
void* microej_malloc(size_t size) {
	void *data;
	int status;

	if (0 == size) {
		return NULL;
	}

	status = tx_byte_allocate(byte_pool, &data, size, TX_NO_WAIT);
	if (TX_SUCCESS != status) {
		return NULL;
	}
	return data;
}

/**
 * @brief Allocate a memory area of n elements of size bytes and return the pointer to the allocated memory.
 * @param[in] n Number of elements to allocate.
 * @param[in] size Number of bytes to allocate per elements.
 * @return A pointer to the allocated memory. NULL if allocation failed.
 */
void* microej_calloc(size_t nmemb, size_t size) {
	return microej_malloc(nmemb * size);
}

/*
* @brief Frees the memory area pointed to by ptr.
* @param[in] ptr Pointer on the memory area which have been returned by a previous call to microej_malloc or microej_calloc.
*/
void microej_free(void *ptr) {
	if(NULL != ptr) {
		tx_byte_release(ptr); // TODO check return value
	}
}
